==> dt-common-0.7.5.tgz: package/battle_engine/ai/dps_ranged/Q_target_high_hp.js <==
const YesNoNode = require('../YesNoNode');
const UseAbility = require('../UseAbility');
const BattleCalc = require('../../BattleCalc');
const Q_most_damage = require('./Q_most_damage');
const { logError } = require('../../logger');

const Q_target_high_hp = function(actor, battleState) {
  const node = new YesNoNode(actor, battleState);

  node.voidSpike = BattleCalc.getEquippedAbility(actor, 'void_spike');
  node.arrowShot = BattleCalc.getEquippedAbility(actor, 'arrow_shot');
  node.razorShot = BattleCalc.getEquippedAbility(actor, 'razor_shot');
  node.infectMind = BattleCalc.getEquippedAbility(actor, 'infect_mind');

  node.branches[ node.NO ] = new Q_most_damage(actor, battleState);

  if (node.razorShot || node.infectMind) {
    node.branches[ node.YES ] = new UseAbility(actor, getBestBleedAbility());
  } else {
    node.branches[ node.YES ] = node.branches[ node.NO ];
  }

  function getBestBleedAbility() {
    var topPotency = node.razorShot? node.razorShot.potency : 0;

    if (node.infectMind && node.infectMind.potency >= topPotency) {
      return 'infect_mind';
    } else if (node.razorShot) {
      return 'razor_shot';
    } else {
      logError(new Error('Attempting bleed ability when none is equipped.'), {
        actor,
        battleState,
      });
      return null;
    }
  }

  node.query = function(target) {
    if (this.yesNoEvalFunction(target)
			&&	(this.razorShot || this.infectMind))
    {
      // for ranged bleeds, try to spread DoTs to non-bleeding targets before applying new stacks
      target = getLeastBleedingTarget(target);

      return this.branches[this.YES].query(target);
    } else {
      return this.branches[this.NO].query([target, this.topDamageHandle]);
    }
  };

  node.yesNoEvalFunction = function(target) {
    var alreadyBleeding = BattleCalc.getCondition(target, 'bleeding');

    if (alreadyBleeding && alreadyBleeding.potency > target.hp*0.25) {
      return false;
    }

    this.topDamageHandle = this.calculateTopDamage();

    return target.hp > this.topDamage*2.5;
  };

  node.calculateTopDamage = function() {
    var voidSpikeDamage = 0;
    var arrowShotDamage = 0;

    if (this.voidSpike) {
      voidSpikeDamage = this.voidSpike.damage;
    }

    if (this.arrowShot) {
      arrowShotDamage = this.arrowShot.damage;
    }

    if (voidSpikeDamage > arrowShotDamage)	{
      this.topDamage = voidSpikeDamage;
      return 'void_spike';
    } else {
      this.topDamage = arrowShotDamage;
      return 'arrow_shot';
    }
  };

  return node;

  function getLeastBleedingTarget(t) {
    let least_stacks = 9999;
    const allies = BattleCalc.getAlliesOfUnit(t, battleState.allUnits);
    for (const ally of allies) {
      const num_bleeds = Object.values(ally.conditions)
        .filter(c => c.handle === 'bleeding')
        .length;
      if (num_bleeds < least_stacks) {
        least_stacks = num_bleeds;
      }
    }

    // now get the closest among those with the least stacks
    let candidates = allies.filter(
      ally => Object.values(ally.conditions).filter(c => c.handle === 'bleeding').length === least_stacks,
    );
    let least_distance = 9999;
    for (const unit of candidates) {
      const dist = BattleCalc.boardDistance(unit, actor);
      if (dist < least_distance) {
        least_distance = dist;
      }
    }
    candidates = candidates.filter(c => BattleCalc.boardDistance(c, actor) === least_distance);
    return candidates[Math.floor(Math.random() * candidates.length)];
  }
};
module.exports = Q_target_high_hp;

==> dt-common-1.0.1.tgz: package/battle_engine/ai/dps_ranged/Q_target_high_hp.js <==
const YesNoNode = require('../YesNoNode');
const UseAbility = require('../UseAbility');
const BattleCalc = require('../../BattleCalc');
const Q_most_damage = require('./Q_most_damage');
const { logError } = require('../../logger');

const Q_target_high_hp = function(actor, battleState) {
  const node = new YesNoNode(actor, battleState);

  node.voidSpike = BattleCalc.getEquippedAbility(actor, 'void_spike');
  node.arrowShot = BattleCalc.getEquippedAbility(actor, 'arrow_shot');
  node.razorShot = BattleCalc.getEquippedAbility(actor, 'razor_shot');
  node.infectMind = BattleCalc.getEquippedAbility(actor, 'infect_mind');

  node.branches[ node.NO ] = new Q_most_damage(actor, battleState);

  if (node.razorShot || node.infectMind) {
    node.branches[ node.YES ] = new UseAbility(actor, getBestBleedAbility());
  } else {
    node.branches[ node.YES ] = node.branches[ node.NO ];
  }

  function getBestBleedAbility() {
    var topPotency = node.razorShot? node.razorShot.potency : 0;

    if (node.infectMind && node.infectMind.potency >= topPotency) {
      return 'infect_mind';
    } else if (node.razorShot) {
      return 'razor_shot';
    } else {
      logError(new Error('Attempting bleed ability when none is equipped.'), {
        actor,
        battleState,
      });
      return null;
    }
  }

  node.query = function(target) {
    if (this.yesNoEvalFunction(target)
			&&	(this.razorShot || this.infectMind))
    {
      // for ranged bleeds, try to spread DoTs to non-bleeding targets before applying new stacks
      target = getLeastBleedingTarget(target);

      return this.branches[this.YES].query(target);
    } else {
      return this.branches[this.NO].query([target, this.topDamageHandle]);
    }
  };

  node.yesNoEvalFunction = function(target) {
    var alreadyBleeding = BattleCalc.getCondition(target, 'bleeding');

    if (alreadyBleeding && alreadyBleeding.potency > target.hp*0.25) {
      return false;
    }

    this.topDamageHandle = this.calculateTopDamage();

    return target.hp > this.topDamage*2.5;
  };

  node.calculateTopDamage = function() {
    var voidSpikeDamage = 0;
    var arrowShotDamage = 0;

    if (this.voidSpike) {
      voidSpikeDamage = this.voidSpike.damage;
    }

    if (this.arrowShot) {
      arrowShotDamage = this.arrowShot.damage;
    }

    if (voidSpikeDamage > arrowShotDamage)	{
      this.topDamage = voidSpikeDamage;
      return 'void_spike';
    } else {
      this.topDamage = arrowShotDamage;
      return 'arrow_shot';
    }
  };

  return node;

  function getLeastBleedingTarget(t) {
    let least_stacks = 9999;
    const allies = BattleCalc.getAlliesOfUnit(t, battleState.allUnits, true);
    for (const ally of allies) {
      const num_bleeds = Object.values(ally.conditions)
        .filter(c => c.handle === 'bleeding')
        .length;
      if (num_bleeds < least_stacks) {
        least_stacks = num_bleeds;
      }
    }

    // now get the closest among those with the least stacks
    let candidates = allies.filter(
      ally => Object.values(ally.conditions).filter(c => c.handle === 'bleeding').length === least_stacks,
    );
    let least_distance = 9999;
    for (const unit of candidates) {
      const dist = BattleCalc.boardDistance(unit, actor);
      if (dist < least_distance) {
        least_distance = dist;
      }
    }
    candidates = candidates.filter(c => BattleCalc.boardDistance(c, actor) === least_distance);
    return candidates[Math.floor(Math.random() * candidates.length)];
  }
};
module.exports = Q_target_high_hp;
